;; Generated package description from vc-hgcmd.el  -*- no-byte-compile: t -*-
(define-package "vc-hgcmd" "1.14" "VC mercurial backend that uses hg command server" '((emacs "25.1")) :keywords '("vc") :authors '(("Andrii Kolomoiets" . "andreyk.mad@gmail.com")) :maintainer '("Andrii Kolomoiets" . "andreyk.mad@gmail.com") :url "https://github.com/muffinmad/emacs-vc-hgcmd")
